#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <sys/stat.h>
#include <string.h>
#include "newsapi.h"
#include "file.h"
#include "util.h"

#ifdef HAVE_READDIR
#include <dirent.h>
#include <ctype.h>

/* based on GRPScandir from inn1.4 */

/* (wish I had something like an iterator)
 *  
 * if ov_list is NULL, it will count either
 *   1) all the articles, or
 *   2) all the articles in the range of [low,high]
 *      if low or high are non-zero.
 * otherwise, it will
 * fill in the article nubers in ov_list.
 */

int scan_group (char * dirname,
		unsigned long * low,
		unsigned long * high,
		unsigned long * count,
		unsigned long artnums[])
{
  struct dirent	*ep;
  DIR		*dp;
  unsigned long	i;
  char *	p;
  unsigned long t_low = ULONG_MAX;
  unsigned long t_high = 0;
  unsigned long index = 0;
  
  if (count)
    *count = 0;

  dp = opendir(dirname);
  if (dp == NULL) {
    return (ERR_OPENING_GROUP);
  }

  while ((ep = readdir(dp)) != NULL) {
    /* Get the numeric value of the filename, if it's all digits. */
    for (p = ep->d_name, i = 0; *p; p++) {
      if (!isdigit(*p))
	break;
      i = i * 10 + *p - '0';
    }
    if (*p || i == 0)
      continue;
    if (i < t_low) t_low = i;
    if (i > t_high) t_high = i;

    if (artnums) {
      /* set the article number */
      artnums[index++] = i;
    }
    else {
      if (*low || *high) {
	if (i >= *low || i <= *high)
	  (*count)++;
      }
      else
	(*count)++;
    }
  }
  closedir(dp);
  if (!artnums) {
    if (!*low && !*high) {
      *low = t_low;
      *high = t_high;
    }
  }
  return (0);
}
#else
int scan_group (char * dirname,
		    unsigned long * low,
		    unsigned long * high,
		    unsigned long * count)
{
  return (ERR_NOT_IMPLEMENTED);
}
#endif

char * construct_group_pathname (news_server_type * server,
				 char * group,
				 unsigned long artnum)
{
  char * path;
  char * group_name;
  char * scan;
  char artnum_string[10];
  file_impdep_type * impdep = FILE_STUFF(server);

  /* copy the group string */
  group_name = string_concatenate (group, NULL);

  if (!group_name) return NULL;

  /* replace the dots with dir_sep */
  for (scan=group_name; *scan ; scan++)
    if (*scan=='.') *scan=*(impdep->dir_sep);
  
  /* if we're looking for an article pathname */
  if (artnum) {
    sprintf (artnum_string, "%lu", artnum);
    path = string_concatenate (impdep->spool_root, 	/* /usr/spool/news */
			       impdep->dir_sep, 	/* / */
			       group_name, 			/* comp/os/linux */
			       impdep->dir_sep, 	/* / */
			       artnum_string, 			/* 123 */
			       NULL);
  }
  else
    path = string_concatenate (impdep->spool_root, /* /usr/spool/news */
			       impdep->dir_sep, /* / */
			       group_name, /* comp/os/linux */
			       NULL);

  news_free (group_name);
  if (!path) return (NULL);
  
  return (path);
}

char * construct_overview_pathname (news_server_type * server,
				    char * group)
{
  char * path;
  char * group_name;
  char * scan;
  file_impdep_type * impdep = FILE_STUFF(server);

  /* copy the group string */
  group_name = string_concatenate (group, NULL);

  if (!group_name) return NULL;

  /* replace the dots with dir_sep */
  for (scan=group_name; *scan ; scan++)
    if (*scan=='.') *scan=*(impdep->dir_sep);
  
  path = string_concatenate (impdep->overview_root, 	/* /usr/spool/news */
			     impdep->dir_sep, 	/* / */
			     group_name, 	/* comp/os/linux */
			     impdep->dir_sep, 	/* / */
			     ".overview",	/* bad, hardcoded */
			     NULL);

  news_free (group_name);
  if (!path) return (NULL);
  
  return (path);
}


#define CHUNK_SIZE	1024
char chunk[CHUNK_SIZE];
char * chunk_index = chunk;

/* slurp a new line from fd into line_buf, safe for obnoxiously
 * long lines
 */

int
slurp_a_line (FILE * fd, buffer_type * line_buf)
{
  int done = 0;
  while (!done) {
    /* if we already have data in the buffer */
    if (chunk_index > chunk) {
      char * nl;
      /* is there a newline in there? */
      nl = (char *) memchr(chunk, '\n', (size_t) (chunk_index - chunk));
    
      if (nl) {
	*nl = '\000';		/* terminate the string */
	
	/* add the contents of the line to the buffer */
	buffer_add_data (line_buf, chunk, (size_t) (nl - chunk + 1) );
	
	nl++;
	
	/* move the leftover data to the front */
	memmove (chunk, nl, (size_t) (chunk_index - nl));
	chunk_index = chunk + (size_t) (chunk_index - nl);
	return (0);
      }
      else {
	/* add the part of the line to the buffer... */
	buffer_add_data(line_buf, chunk, (size_t) (chunk_index - chunk));
	chunk_index = chunk;
	/* and try again */
      }
    }
    /* we need to grab some data ... */
    else {
      size_t num_read = fread(chunk_index, (size_t)1, (size_t)CHUNK_SIZE, fd);
      if (num_read)
	chunk_index = chunk_index + num_read;
      else
	return (-1);
    }
  }
  /* shouldn't get here */
  return (-1);
}  

typedef char * overview_temp[5];

char * overview_headers[] = {   "Subject: ",
				"From: ",
				"Date: ",
				"Message-ID: ",
				"References: " };

int
generate_overview (FILE * fd,
		   unsigned long artnum,
		   news_overview_type * nov,
		   buffer_type * line_buf,
		   buffer_type * ov_buf)
{
  overview_temp have = {NULL,NULL,NULL,NULL,NULL};
  int overview_lens[5] = {9,6,6,12,12};
  int i;
  int last_head = -1;
  unsigned int lines = 0;	/* FIXME (fill me in!) */
  unsigned int bytes = 0;	/* FIXME (fill me in!) */
  
  if (fd) {
    while (slurp_a_line(fd, line_buf) != -1) {
      /* blank line? (end of headers?) */
      if (!lines && line_buf->data[0] != '\000') {
	/* is this a continuation line? */

	if (isspace(line_buf->data[0])) {
	  /* are we expecting a continuation? */
	  if (last_head >= 0) {
	    char * p = line_buf->data;
	    /* skip past the whitespace */
	    while (isspace(*p))
	      p++;
	    /* back up to ignore the NULL */
	    ov_buf->index--;
	    buffer_add_data (ov_buf, p, strlen(p) + 1);
	    /* add a space */
	    buffer_add_data (ov_buf, " \000", 2);
	  }
	}

	/* a new header */

	else {
	  /* do we want this header */
	  for (i = 0; i < 5; i++) {
	    if (strncmp(overview_headers[i], line_buf->data, overview_lens[i]) == 0) {
	      char * skip = line_buf->data + overview_lens[i];
	      /* store this string away */
	      have[i] = ov_buf->index;
	      buffer_add_data (ov_buf,
			       skip,
			       strlen(skip) + 1);
	      /* keep track of which header in case of continuation lines */
	      last_head = i;
	    }
	  }
	}
      }
      else {
	lines++;
      }
      buffer_reset (line_buf);
    }
    /* now we have the data, stored into ov_buf */
    nov->subject =	have[0];
    nov->from =		have[1];
    nov->date =		parsedate (have[2], (TIMEINFO *) NULL);
    nov->message_id =	have[3];
    nov->references =	have[4];
    nov->lines =	lines;
    nov->bytes =	bytes;
    nov->number =	artnum;
    return (0);
  }
  return (-1);
}





/* filling in overview_list, the plan:
 * find out what articles are on file, make an overview_list
 * to hold each (fill in artnums).
 * suck up the .overview into a buffer
 * iterate parse_overview on the relevant lines
 * make another pass, making calls to generate_overview to
 * fill in the missing ones...
 */


